@@ -29,6 +29,37 @@
 #include <apr_pools.h>
 #include <apr_tables.h>
 
+#define REDWAX_DECLARE_MODULE(foo) \
+    module foo##_module
+
+/** Use this in all standard modules */
+#define STANDARD_MODULE_STUFF -1, \
+                              __FILE__, \
+                                NULL
+
+typedef struct module_struct module;
+struct module_struct {
+    int module_index;
+    const char *name;
+    struct module_struct *next;
+    void (*register_hooks) (apr_pool_t *p);
+};
+
+typedef struct redwax_conf_vector_t redwax_conf_vector_t;
+
+redwax_conf_vector_t *redwax_create_module_config(apr_pool_t *p);
+
+void *redwax_get_module_config(const redwax_conf_vector_t *cv,
+                                        const module *m);
+
+void redwax_set_module_config(redwax_conf_vector_t *cv, const module *m,
+                                      void *val);
+
+#define redwax_get_module_config(v,m)       \
+    (v ? ((void **)(v))[(m)->module_index] : NULL)
+#define redwax_set_module_config(v,m,val)   \
+    ((((void **)(v))[(m)->module_index]) = (val))
+
 typedef struct redwax_filter_t {
     int filter_applied;
 } redwax_filter_t;
@@ -69,6 +100,7 @@ typedef enum redwax_expiry_e {
 typedef struct redwax_tool_t {
     apr_pool_t *pool;
     apr_pool_t *tpool;
+    redwax_conf_vector_t *per_module;
     apr_file_t *err;
     apr_file_t *in;
     apr_file_t *out;
@@ -94,6 +126,7 @@ typedef struct redwax_tool_t {
     apr_hash_t *ips_index;
     apr_hash_t *keys_index;
     apr_hash_t *duplicates_index;
+    apr_hash_t *trust_duplicates_index;
     apr_hash_t *cert_relationships;
     const char *verify_param;
     const char *verify_date;
@@ -191,6 +224,7 @@ typedef struct redwax_certificate_x509_t {
 typedef struct redwax_certificate_t {
     apr_pool_t *pool;
     struct redwax_certificate_t *next;
+    redwax_conf_vector_t *per_module;
     const char *header;
     const unsigned char *der;
     apr_size_t len;
@@ -271,6 +305,7 @@ typedef struct redwax_key_rsa_t {
 
 typedef struct redwax_key_t {
     apr_pool_t *pool;
+    redwax_conf_vector_t *per_module;
     const char *header;
     const unsigned char *der;
     apr_size_t len;
@@ -412,6 +447,14 @@ APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, initialise,
 APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_pem_in,
         (redwax_tool_t *r, const char *arg, const char *secret));
 
+/**
+ * Hook to process incoming trusted certificates.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_trust_pem_in,
+        (redwax_tool_t *r, const char *arg, const char *secret));
+
 /**
  * Hook to complete PKCS11 URL in.
  *
@@ -444,6 +487,22 @@ APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, complete_pkcs11_module_in,
 APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_pkcs11_module_in,
         (redwax_tool_t *r, const char *arg));
 
+/**
+ * Hook to complete keychain in.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, complete_keychain_in,
+        (redwax_tool_t *r, const char *url, apr_hash_t *urls));
+
+/**
+ * Hook to write incoming keychain entries.
+ *
+ * @param r The redwax-tool context.
+ */
+APR_DECLARE_EXTERNAL_HOOK(rt, REDWAX, apr_status_t, process_keychain_in,
+        (redwax_tool_t *r, const char *arg));
+
 /**
  * Hook to complete incoming certificates / intermediates / keys
  * and pass filtered results to the outgoing certificates /
